/*
  File name        : cycle.c
  Date of creation : 1/5/2009
  Version          : 1.0
  Copyright        : Soft'n'design
  Author           : Laurent Mazet <mazet@softndesign.org>

  Description      : This file contains functions for profiling

  History          :
  - initial version
*/

/* depend: */
/* cflags: */
/* lflags: */

#include <assert.h>
#include <errno.h>
#include <inttypes.h>
#include <malloc.h>
#include <math.h>
#include <pthread.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <sys/syscall.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>

/**
    Wrapper for gettid
*/
inline pid_t gettid (void)
{
    return syscall (SYS_gettid);
}

/**
   Cycle counter control
*/
int enable_cycle_counters_ = -1;

/**
   Cpu clock
*/
double cpu_clock = 1;

#include "cycle.h"

/**
   Check environment variable CYCLE_COUNTERS_ to set cycle counter
   control. Possible values are (0, "no") or (1, "yes").
*/
int enable_cycle_counters (void)
{
    if (enable_cycle_counters_ < 0) {
        char *answer = getenv ("CYCLE_COUNTERS_");
        enable_cycle_counters_ = (answer == NULL) ? 0 : atoi (answer);
        if (enable_cycle_counters_ <= 0) {
            char test = (answer == NULL) ? 0 : answer[0];
            switch (test) {
            case 'Y': /* yes */
            case 'y':
                enable_cycle_counters_ = 1;
                break;
            default:
                enable_cycle_counters_ = 0;
            }
        }

        /* estimate cpu clock */
        if (enable_cycle_counters_) {
            struct timeval tv1, tv2;
            int64_t ts1, ts2;

            gettimeofday (&tv1, NULL);
            ts1 = rdtsc ();

            usleep (100000);

            gettimeofday (&tv2, NULL);
            ts2 = rdtsc ();

            double delta = ((double)tv2.tv_sec + tv2.tv_usec * 1e-6) -
                ((double)tv1.tv_sec + tv1.tv_usec * 1e-6);

            cpu_clock = (ts2 - ts1) / delta;
        }
    }

    return enable_cycle_counters_;
}

/**
   Gobal structure which store all cycle counters.
*/
static cycle_counter_t **cycle_counters = NULL;

/**
   Global number of cycle counters.
*/
static int nb_counters = 0;

/**
   Thread protection mutex
*/
static pthread_mutex_t cycle_counter_mutex = PTHREAD_MUTEX_INITIALIZER;

cycle_counter_t *create_cycle_counter (const char *file, const char *function)
{

    /* thread protection begin */
    pthread_mutex_lock (&cycle_counter_mutex);

    /* allocatate memory for new counter */
    cycle_counters = (cycle_counter_t **)
        realloc (cycle_counters, ++nb_counters  * sizeof (cycle_counter_t*));
    assert (cycle_counters);
    cycle_counter_t *counter = cycle_counters[nb_counters - 1] =
        (cycle_counter_t *) malloc (sizeof (cycle_counter_t));
    assert (counter);
    counter->last_nb_cycles =
        (int64_t *) calloc (enable_cycle_counters_, sizeof (int64_t));
    assert (counter->last_nb_cycles);

    /* thread protection end */
    pthread_mutex_unlock (&cycle_counter_mutex);

    /* fill counter fields */
    counter->file = strdup (file);
    counter->function = strdup (function);
    counter->tid = gettid ();
    counter->ptid = pthread_self ();
    counter->nb_cycles = 0;
    counter->variance = 0;
    counter->nb_calls = 0;
    counter->last = 0;

    return counter;
}

float _sqrtf (float val)
{
    /* this function 5 to 10 times slow than standard sqrtf but it's
     * avoid dependency to math library */

    float res = 0;

    /* check error */
    if ((isnan (val)) || (val < 0)) {
        errno = EDOM;
        return NAN;
    }
    if (!isfinite (val))
        return val;

    /* compute magnitude */
    float x = val;
    int n = 1;
    if (x > 10) {
        while (x > 1) {
            n++;
            x = x / 10;
        }
        n--;
    } else if (x < 1) {
        while (x < 1) {
            n--;
            x = x * 10;
        }
    }

    /* compute incremental step */
    n = n / 2 + ((n < 0) ? -1 : (n > 0) ? 1 : 0);
    float step = 1;
    if (n < 0) {
        while (n < 0) {
            step /= 10;
            n++;
        }
    } else {
        while (n > 0) {
            step *=10;
            n--;
        }
    }

    /* compute square root */
    float tmp = 0;
    while (tmp * tmp <= val)
        tmp += step;

    int j;
    for (j = 0; j < 10; j++)
        tmp = res = (val / tmp + tmp) / 2;

    return res;
}

void display_cycle_counters (void)
{
    cycle_counter_t *current;
    int i, j;

    /* check that cylce counters have been used */
    if (!nb_counters)
        return;

    /* sort ounter list */
    for (i = 0; i < nb_counters; i++)
        for (j = 1; j < nb_counters; j++)
            if (cycle_counters[j - 1]->nb_cycles < cycle_counters[j]->nb_cycles) {
                cycle_counter_t *tmp = cycle_counters[j - 1];
                cycle_counters[j - 1] = cycle_counters[j];
                cycle_counters[j] = tmp;
            }

    /* display table header */
    printf ("cycle counters summary (CPU clock %.3fGHz)\n",
            cpu_clock / 1e9f);
    printf ("-------------------------------------------\n");
    printf ("  total cycles | calls | cyc./call  us (std) | function name [file name] (tid/ptid)");
    if (enable_cycle_counters_ > 1)
        printf (" last %d cyc./cal (us) measurements", enable_cycle_counters_);
    printf ("\n");

    float delta = 1e6f / cpu_clock;
    for (i = 0; i < nb_counters; i++) {
        current = cycle_counters[i];
        if (current->nb_calls == 0)
            continue;
        if (current->nb_calls > 1) {
            float mean = current->nb_cycles / (float)current->nb_calls;
            float std_dev = _sqrtf ((current->variance - mean * mean * current->nb_calls) / (float)(current->nb_calls - 1));
            int is_mesurable = (current->nb_cycles * delta) / current->nb_calls > 1;
            printf ("%15"PRId64" %7d %10"PRId64"%c%10.1f(%.1f) %s [%s] (%u/0x%lx)",
                    current->nb_cycles, current->nb_calls,
                    current->nb_cycles / current->nb_calls,
                    (is_mesurable) ? ' ' : '!',
                    mean * delta, std_dev * delta,
                    current->function, current->file, current->tid, current->ptid);
        } else
            printf ("%15"PRId64" %7d          - %10.1f  %s [%s] (%u/0x%lx)",
                    current->nb_cycles, current->nb_calls,
                    current->nb_cycles * delta,
                    current->function, current->file, current->tid, current->ptid);
        if (enable_cycle_counters_ > 1) {
            int first = current->nb_calls - enable_cycle_counters_;
            for (j = (first > 0) ? first : 0; j < current->nb_calls; j++)
                printf (" %.1f", current->last_nb_cycles[j % enable_cycle_counters_] * delta);
        }
        printf ("\n");
    }
}

/* vi:set tabstop=4 expandtab shiftwidth=4: this line set vi mode */
